/***********************************************************************
ColorSafe 1.0 Beta - http://www.woven-media.com/

A Java 6 Swing application for color testing and conversions from RGB to
hexadecimal data formats and vice versa.

Copyright (C) 2008-2009, Brent Allen Parrish <info@woven-media.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
************************************************************************/
package com.woven_media.colorsafe;

import java.awt.Color;
import java.util.StringTokenizer;
import java.text.NumberFormat;
import java.text.DecimalFormat;


/**
 * ColorUtil is a static class for calculating color conversions and color models.
 *
 * @author  Brent Allen Parrish
 * @version %I%, %G%
 * @since   1.0
 */
public class ColorUtil
{
    /**
     * Converts RGB integer values to a hexadecimal triplet and return as
     * <code>java.lang.String</code>.
     *
     * @param red   <code>int</code> value representing red.
     * @param green <code>int</code> value representing green.
     * @param blue  <code>int</code> value representing blue.
     * @since       1.0
     */
    public static String RGBtoHEX(int red, int green, int blue)
    {
        String hex[] = {"0","1","2","3","4","5","6","7","8","9","A","B","C","D","E","F"};
        String r = hex[red   / 16] + hex[red   % 16];
        String g = hex[green / 16] + hex[green % 16];
        String b = hex[blue  / 16] + hex[blue  % 16];
        String hexValue = r + g + b;

        return hexValue;
    }

    /**
     * Tests whether a string is a valid hexadecimal value.
     *
     * @param  hex string representing a hexadecimal triplet number.
     * @return     <code>boolean</code> value indicating a valid or invalid hexadecimal string.
     * @since      1.0
     */
    public static boolean isValidHex(String hex)
    {
        return (isValidString(hex, "0123456789ABCDEF")) ? true : false;
    }

    /**
     * Tests whether a string contains the required characters.
     *
     * @param  inString     string to check.
     * @param  filterString string of required characters.
     * @return              <code>boolean</code> value indicating a valid or invalid string.
     * @since               1.0
     */
    public static boolean isValidString(String inString, String filterString)//TODO: This belongs in String class!
    {
        boolean result = false;
        if((inString.length() > 0 || inString != null))
        {
            int ctr = 0;

            String inStr = inString.toUpperCase();
            String workStr = filterString;
            char[] chars = inStr.toCharArray();
            char[] filterChars = workStr.toCharArray();

            for(int i=0; i < chars.length; i++)
            {
                for(int j=0; j < filterChars.length; j++)
                {
                    if(chars[i] == filterChars[j])
                    {
                        ctr++;
                    }
                }
            }
            result = (ctr == chars.length) ? true : false;
        }
        return result;
    }

    /**
     * Returns a RGB <code>java.awt.Color</code> object by parsing the <code>hex</code> string
     * parameter. An abbreviated, three (hexadecimal) digit form can be used, as well as the
     * standard six-digit form.
     * <h4>e.g.</h4>
     * 09C becomes 0099CC<br>
     * FFF becomes FFFFFF<br>
     * CDA becomes CCDDAA
     *
     * @param  hex string representing a hexadecimal triplet or abbreviated triplet.
     * @return     <code>java.awt.Color</code> object created from the <code>hex</code> string equivalent.
     * @see        ColorUtil#parseHex
     * @since      1.0
     */
    public static Color getRGBColor(String hex)
    {
        Color rgb = null;
        if(isValidHex(hex))
        {
            try
            {
                String[] hexs = parseHex(hex);
                int r = Integer.parseInt(hexs[0], 16);
                int g = Integer.parseInt(hexs[1], 16);
                int b = Integer.parseInt(hexs[2], 16);
                rgb = new Color(r, g, b);
            }
            catch(NumberFormatException nfe)
            {
                return rgb;
            }
        }
        return rgb;
    }

    /**
     * Returns a <code>java.lang.String</code> array of the seperate red, green
     * and blue hexadecimal values and checks for the abbreviated form. If the
     * <code>hex</code> parameter is an abbreviated form, a six-digit standard form
     * hexadecimal triplet is created and returned as an array of strings.
     *
     * @param hex hexadecimal triplet string, either six-digit or three-digit form.
     * @return    string array of the seperate red, green and blue hexadecimal triplet values,
     *            or an empty array if an invalid hex parameter is passed to the function.
     * @see       ColorUtil#getRGBColor
     * @since     1.0
     */
    public static String[] parseHex(String hex)
    {
        int len = hex.length();
        String redstr = "00";
        String grnstr = "00";
        String blustr = "00";
        String[] hexs = {"00", "00", "00"}; // default
        if(len == 3) // Abbreviated form
        {
            redstr = hex.substring(0,1); redstr += redstr;
            grnstr = hex.substring(1,2); grnstr += grnstr;
            blustr = hex.substring(2,3); blustr += blustr;
        }
        else if(len == 6)
        {
            redstr = hex.substring(0,2);
            grnstr = hex.substring(2,4);
            blustr = hex.substring(4,6);
        }
        hexs[0] = redstr;
        hexs[1] = grnstr;
        hexs[2] = blustr;

        return hexs;
    }

    /**
     * Returns an array of Arithmetic (floating point values) RGB values.
     *
     * @param  hex string representing hexadecimal RGB value.
     * @return     3-element floating point array of Arithmetic RGB values.
     * @see        java.lang.Integer
     * @since      1.0
     */

    public static double[] hexToFloatArray(String hex)
    {
        String hexValue = (isValidHex(hex)) ? hex : "000";
        int rgb = Integer.parseInt(hexValue, 16);
        double red   = ((rgb >> 16) & 0xFF);
        double green = ((rgb >> 8) & 0xFF);
        double blue  = (rgb & 0xFF);
        red   = roundThreeDecimals((red / 255.0));
        green = roundThreeDecimals((green / 255.0));
        blue  = roundThreeDecimals((blue / 255.0));
        double[] colorArray = {red, green, blue};

        return colorArray;
    }

    /**
     * Returns an array of Decimal (integer values) RGB values.
     *
     * @param  r double representing red arithmetic RGB value.
     * @param  g double representing green arithmetic RGB value.
     * @param  b double representing blue arithmetic RGB value.
     * @return   3-element integer array of Decimal RGB values.
     * @see      java.lang.Double
     * @since    1.0
     */
    public static int[] rgbFloatToUInt(double r, double g, double b)
    {
        int red = scaleToUInt(r);
        int grn = scaleToUInt(g);
        int blu = scaleToUInt(b);
        int[] rgb = {red, grn, blu};

        return rgb;
    }

    /**
     * Converts a double arithmetic RGB value to an unsigned integer value.
     *
     * @param  d double representing arithmetic RGB value.
     * @return   rounded int value.
     * @since    1.0
     */
    public static int scaleToUInt(double d)
    {
       return (int)Math.round(d * 0xff);
    }

    /**
     * Formats and rounds a <code>java.lang.Double</code> type number to the third decimal place.
     *
     * @param   d <code>java.lang.Double</code> floating point number.
     * @return    <code>java.lang.Double</code> type number rounded the third decimal place.
     * @see       java.lang.Double
     * @since     1.0
     */
    public static double roundThreeDecimals(double d)
    {
        DecimalFormat threeDForm = new DecimalFormat("#.###");
        return Double.valueOf(threeDForm.format(d));
    }

    public static String removeTrailingZeros(String str)
    {
        if (str == null)
        {
            return null;
        }
        char[] chars = str.toCharArray();
        int length,index;
        length = str.length();
        index = length -1;
        for(; index >=0;index--)
        {
            if (chars[index] != '0')
            {
                break;
            }
        }
        return (index == length-1) ? str : str.substring(0, index+1);
    }

    /**
     * Returns a <code>java.awt.Color</code> object from RGB strings.
     *
     * @param  red string representing red.
     * @param  grn string representing green.
     * @param  blu string representing blue.
     * @return     <code>java.awt.Color</code> object created from RGB string values.
     * @see        java.lang.NumberFormatException
     * @see        java.lang.Integer
     * @since      1.0
     */
    public static Color getRGBColor(String red, String grn, String blu)
    {
        Color rgb = null;
        try
        {
            int r = (red.length() != 0) ? Integer.parseInt(red) : 0;
            int g = (grn.length() != 0) ? Integer.parseInt(grn) : 0;
            int b = (blu.length() != 0) ? Integer.parseInt(blu) : 0;
            rgb = new Color(r, g, b);
        }
        catch(NumberFormatException nfe)
        {
            return rgb;
        }
        return rgb;
    }

    /**
     * Determines whether a hexadecimal string is a safe color equivalent.
     *
     * @param  hex string representing a hexadecimal triplet number.
     * @return     <code>boolean</code> value indicating whether a color is a safe color or not.
     * @since      1.0
     */
    public static boolean isColorSafe(String hex)
    {
        boolean result;
        String red, green, blue;
        int r, g, b;
        result = false;

        if((hex.length() == 6) && isValidHex(hex))
        {
            red = hex.substring(0,2);
            green = hex.substring(2,4);
            blue = hex.substring(4,6);
            r = HEXtoRGB(red);
            g = HEXtoRGB(green);
            b = HEXtoRGB(blue);

            if(((r % 51) == 0) && ((g % 51) == 0) && ((b % 51) == 0))
            {
                result = true;
            }
        }
        return result;
    }

    /**
     * Builds an array of RGB values based on a hexadecimal triplet.
     *
     * @param  hex string representing a hexadecimal triplet number.
     * @return     <code>int</code> array of RGB values.
     * @since      1.0
     */
    public static int[] buildRGBarray(String hex)
    {
        int r, g, b;
        int[] rgb = new int[3];
        String red, grn, blu;
        String[] hexs = new String[3];
        hexs = parseHex(hex); // Check for abbreviated form
        red = hexs[0];
        grn = hexs[1];
        blu = hexs[2];
        r = HEXtoRGB(red);
        g = HEXtoRGB(grn);
        b = HEXtoRGB(blu);
        rgb[0] = r;
        rgb[1] = g;
        rgb[2] = b;

        return rgb;
    }

    /**
     * Algorithm for determining the nearest safe color based on the
     * values of a valid RGB array of unsigned integer values.
     *
     * @param  rgbArray <code>int</code> array of representing RGB values.
     * @return          <code>int</code> array of RGB values converted to a safe color equivalent.
     * @since           1.0
     */
    public static int[] nearestSafeColor(int[] rgbArray)
    {
        int q;
        int[] rgb = new int[3];

        for(int i=0; i < rgbArray.length; i++)
        {
            q = (rgbArray[i] / 51);

            if((rgbArray[i] % 51) != 0)
            {
                switch(q)
                {
                case 0: // CASE 0: under 51 > 0
                    if(rgbArray[i] < 26) {rgb[i] = 0;}
                    else {rgb[i] = 51;}
                    break;
                case 1: // CASE 1: under 102 > 51
                    if(rgbArray[i] < 77) {rgb[i] = 51;}
                    else {rgb[i] = 102;}
                    break;
                case 2: // CASE 2: under 153 > 102
                    if(rgbArray[i] < 128) {rgb[i] = 102;}
                    else {rgb[i] = 153;}
                    break;
                case 3: // CASE 3: under 204 > 153
                    if(rgbArray[i] < 179) {rgb[i] = 153;}
                    else {rgb[i] = 204;}
                    break;
                case 4: // CASE 4: under 255 > 204
                    if(rgbArray[i] < 230) {rgb[i] = 204;}
                    else {rgb[i] = 255;}
                    break;
                }
            }
            else
            {
                rgb[i] = rgbArray[i];
            }
        }

        return rgb;
    }

    /**
     * Approximation algorithm only for CMYK color! Converts RGB
     * values to CMYK array of values. Values are between 0 - 255.
     *
     * @param  red <code>int</code> value between (0-255) representing red.
     * @param  grn <code>int</code> value between (0-255) representing green.
     * @param  blu <code>int</code> value between (0-255) representing blue.
     * @return     <code>int</code> array of CMYK values converted from RGB.
     * @since      1.0
     */
    public static int[] rgbToCMYK(int red, int grn, int blu)
    {
        int C, M, Y, K;
        int[] cmyk = new int[4];
        //RGB to CMY
        C = 255 - red;
        M = 255 - grn;
        Y = 255 - blu;
        //CMY to CMYK
        K = Math.min(C, Math.min(M, Y));
        cmyk[0] = (C - K); // Cyan
        cmyk[1] = (M - K); // Magenta
        cmyk[2] = (Y - K); // Yellow
        cmyk[3] = K;       // Black

        return cmyk;
    }

    /**
     * Parses a <code>java.awt.Color</code> object and returns an array of RGB integers.
     *
     * @param  c <code>java.awt.Color</code> object.
     * @return   <code>int</code> array of RGB values converted from a <code>java.awt.Color</code> object.
     * @see      java.util.StringTokenizer
     * @see      java.lang.StringBuffer
     * @since    1.0
     */
    public static int[] parseAWTColor(Color c)
    {
        int i, r, g, b;
        int[] rgb;
        String s;
        String[] temp;
        StringBuffer[] arr;
        StringTokenizer st;
        StringBuffer wrkStr = new StringBuffer(c.toString());

        wrkStr.delete(0, 15);
        wrkStr.deleteCharAt(wrkStr.length()-1);
        s = wrkStr.toString();

        st = new StringTokenizer(s, ",");
        arr = new StringBuffer[3];
        temp = new String[3];
        rgb = new int[3];

        i = 0;
        while(st.hasMoreTokens())
        {
            try
            {
                temp[i] = st.nextToken(",");
                arr[i] = new StringBuffer(temp[i]);
                temp[i] = arr[i].delete(0, 2).toString();
                rgb[i] = Integer.parseInt(temp[i]);
                i++;
            }
            catch(NumberFormatException ignored)
            {
            }
        }
        return rgb;
    }

    /**
     * Converts CMYK values between 0 - 255 to values
     * between 0 - 100 that can be expressed as percentages.
     *
     * @param  c <code>int</code> representing cyan.
     * @param  m <code>int</code> representing magenta.
     * @param  y <code>int</code> representing yellow.
     * @param  k <code>int</code> representing black.
     * @return   <code>double</code> array of CMYK percentages.
     * @see      java.text.NumberFormat
     * @since    1.0
     */
    public static double[] cmykAsPercent(int c, int m, int y, int k)
    {
        double C, M, Y, K;
        double[] cmyk = new double[4];

        NumberFormat f = NumberFormat.getInstance();

        f.setMaximumFractionDigits(0);

        C = Math.abs((c / 255.0) * 100);
        M = Math.abs((m / 255.0) * 100);
        Y = Math.abs((y / 255.0) * 100);
        K = Math.abs((k / 255.0) * 100);

        cmyk[0] = (Double.parseDouble(f.format(C)) / 100) * 100; // Cyan
        cmyk[1] = (Double.parseDouble(f.format(M)) / 100) * 100; // Magenta
        cmyk[2] = (Double.parseDouble(f.format(Y)) / 100) * 100; // Yellow
        cmyk[3] = (Double.parseDouble(f.format(K)) / 100) * 100; // Black

        return cmyk;
    }

    /**
     * Converts CMYK values to RGB.
     *
     * @param  c <code>int</code> representing cyan.
     * @param  m <code>int</code> representing magenta.
     * @param  y <code>int</code> representing yellow.
     * @param  k <code>int</code> representing black.
     * @return   <code>int</code> array of RGB values converted from CMYK.
     * @since    1.0
     */
    public static int[] cmykToRGB(int c, int m, int y, int k)
    {
        int R, G, B;
        int[] rgb = new int[3];

        rgb[0] = 255 - (c + k); // Red
        rgb[1] = 255 - (m + k); // Green
        rgb[2] = 255 - (k + k); // Blue

        return rgb;
    }

    /**
     * Get HSB values from RGB color model.
     *
     * @param  r <code>int</code> representing red.
     * @param  g <code>int</code> representing green.
     * @param  b <code>int</code> representing blue.
     * @return   string array of HSB values converted from RGB.
     * @since    1.0
     */
    public static String[] getHSBvalues(int r, int g, int b)
    {
        String[] HSBvals = new String[3];
        float[] HSB = Color.RGBtoHSB(r, g, b, (new float[3]));
        HSB[0] *= 360;
        HSB[1] *= 100;
        HSB[2] *= 100;
        HSBvals[0] = String.valueOf((int)HSB[0]);
        HSBvals[1] = String.valueOf((int)HSB[1]);
        HSBvals[2] = String.valueOf((int)HSB[2]);

        return HSBvals;
    }

    /**
     * Returns a <code>java.awt.Color</code> object representing the nearest safe color.
     *
     * @param  rgbArray  array of valid RGB values.
     * @return           <code>java.awt.Color</code> object
     * @since            1.0
     */
    public static Color getNearestSafeColor(int[] rgbArray)
    {
        Color safeColor;
        int[] rgb = nearestSafeColor(rgbArray);
        safeColor = new Color(rgb[0], rgb[1], rgb[2]);

        return safeColor;
    }

    /**
     * Convenience routine that converts individual triplet values of
     * a hexadecimal string to an integer.
     * e.g.
     * HEXtoRGB(&quot;FF&quot;) not HEXtoRGB(&quot;FFFFFF&quot;)
     *
     * @param  hexPart string representing a single hexadecimal triplet number.
     * @return         <code>int</code> value of either red, green or blue.
     * @since          1.0
     */
    public static int HEXtoRGB(String hexPart)
    {
        int retVal;

        try
        {
            retVal = Integer.parseInt(hexPart, 16);
        }
        catch(NumberFormatException nfe)
        {
            retVal = 0;
        }

        return retVal;
    }

    /**
     * Converts strings representing Red RGB values to integers.
     *
     * @param  red string representing red
     * @return     <code>int</code> value representing red converted from a string.
     * @since      1.0
     */
    public static int getRed(String red)
    {
        int r;
        try
        {
            r = Integer.parseInt(red);
        }
        catch(NumberFormatException nfe)
        {
            r = 0;
        }
        return r;
    }

    /**
     * Converts strings representing Green RGB values to integers.
     *
     * @param  green string representing green.
     * @return       <code>int</code> value representing green converted from a string.
     * @since        1.0
     */
    public static int getGreen(String green)
    {
        int g;
        try
        {
            g = Integer.parseInt(green);
        }
        catch(NumberFormatException nfe)
        {
            g = 0;
        }
        return g;
    }

    /**
     * Converts strings representing Blue RGB values to integers.
     *
     * @param  blue string representing blue.
     * @return      <code>int</code> value representing blue converted from a string.
     * @since       1.0
     */
    public static int getBlue(String blue)
    {
        int b;
        try
        {
            b = Integer.parseInt(blue);
        }
        catch(NumberFormatException nfe)
        {
            b = 0;
        }
        return b;
    }

    /**
     * Grab the luminance the same way GIMP does.
     * <p>i.e. <code>Math.round(0.3 * r + 0.59 * g + 0.11 * b)</code></p>
     *
     * @return <code>int</code> value representing the luminance.
     * @since  1.0
     */
    public int luminance(float r, float g, float b)
    {
        return (int) Math.round(0.3 * r + 0.59 * g + 0.11 * b);
    }

    /**
     * Converts to HSV color model from RGB model.
     * <ul>
     * <li>r,g,b values are from 0 to 1</li>
     * <li>h = [0,360], s = [0,1], v = [0,1]</li>
     * <li>if s == 0, then h = -1 (undefined)</li>
     * </ul>
     *
     * @param  r   represents red as a float value        [0 - 1]
     * @param  g   represents green as a float value      [0 - 1]
     * @param  b   represents blue as a float value       [0 - 1]
     * @param  h   represents hue as a float value        [0,360]
     * @param  s   represents saturation as a float value [0,1]
     * @param  v   represents value as a float value      [0,1]
     * @return     <code>float</code> array of HSV values converted from RGB.
     * @since      1.0
     */
    public static float[] RGBtoHSV( float r, float g, float b, float h, float s, float v )
    {
        float min, max, delta;
        float[] hsv = new float[3];     // return value

        min = Math.min(r, g);
        min = Math.min(min, b);
        max = Math.max(r, g);
        max = Math.max(max, b);
        v = max;                        // v

        delta = max - min;

        if(max != 0)
        {
            s = delta / max;            // s
        }
        else
        {
            // r = g = b = 0            // s = 0, v is undefined
            s = 0;
            h = -1;
            return new float[0];
        }
        if(r == max)
        {
            h = (g - b) / delta;        // between yellow & magenta
        }
        else if(g == max)
        {
            h = 2 + (b - r) / delta;    // between cyan & yellow
        }
        else
        {
            h = 4 + (r - g) / delta;    // between magenta & cyan
        }
        h *= 60;                        // degrees
        if(h < 0)
        {
            h += 360;
        }

        hsv[0] = h;                     // Assign return values
        hsv[1] = s;
        hsv[2] = v;

        return hsv;
    }


    /**
     * Converts to RGB color model from HSV model.
     *
     * @param  r   represents red as a float value        [0 - 1]
     * @param  g   represents green as a float value      [0 - 1]
     * @param  b   represents blue as a float value       [0 - 1]
     * @param  h   represents hue as a float value        [0,360]
     * @param  s   represents saturation as a float value [0,1]
     * @param  v   represents value as a float value      [0,1]
     * @return     <code>float</code> array of RGB values converted from HSV.
     * @since      1.0
     */
    public static float[] HSVtoRGB(float r, float g, float b, float h, float s, float v)
    {
        int i;
        float f, p, q, t;
        float[] rgb = new float[3];        // return value

        if(s == 0)
        {
            // achromatic (grey)
            r = g = b = v;
            return new float[0];
        }

        h /= 60;                          // sector 0 to 5
        i = (int) Math.floor((double) h);
        f = h - i;                        // factorial part of h
        p = v * (1 - s);
        q = v * (1 - s * f);
        t = v * (1 - s * (1 - f));

        switch(i)
        {
        case 0:
            r = v;
            g = t;
            b = p;
            break;
        case 1:
            r = q;
            g = v;
            b = p;
            break;
        case 2:
            r = p;
            g = v;
            b = t;
            break;
        case 3:
            r = p;
            g = q;
            b = v;
            break;
        case 4:
            r = t;
            g = p;
            b = v;
            break;
        default:        // case 5:
            r = v;
            g = p;
            b = q;
            break;
        }

        rgb[0] = r;
        rgb[1] = g;
        rgb[2] = b;

        return rgb;
    }

    /* ######### FOR DEBUG ONLY #########
    public static void main(String[] args)
    {
        System.out.println();
    }
    */
} // end class
